from typing import Optional

import numpy as np

from barrage import api


class KeySelector(api.RecordLoader):
    """Record loader for directly transforming keys from a record into a data record.

    Args:
        mode: RecordMode, load mode.
        inputs: dict, {input: key or [keys], ...}.
        outputs: dict, {output: key or [keys], ...}.
        sample_weights: dict or None (OPTIONAL), {output: key, ...}.

    Raises:
        TypeError, invalid params.
    """

    def __init__(
        self,
        mode: api.RecordMode,
        inputs: dict,
        outputs: dict,
        sample_weights: Optional[dict] = None,
    ):
        super().__init__(mode)

        self.inputs = inputs
        self.outputs = outputs
        self.sample_weights = sample_weights

        if not isinstance(self.inputs, dict):
            raise TypeError("KeySelector param 'inputs' must be type dict")
        if not isinstance(self.outputs, dict):
            raise TypeError("KeySelector param 'outputs' must be type dict")

        if not (isinstance(self.sample_weights, dict) or self.sample_weights is None):
            raise TypeError("KeySelector 'sample_weights' must be type dict or None")

    def load(self, record: api.Record) -> api.DataRecord:
        """Load a record by selecting keys corresponding to inputs, outputs, and
        maybe sample weights.

        Args:
            record: Record, record.

        Returns:
            DataRecord, data record.
        """

        def _index_dict_to_arr(d, keys):
            if isinstance(keys, list):
                return np.array([d[k] for k in keys])
            else:
                return np.array(d[keys])

        X = {k: _index_dict_to_arr(record, v) for k, v in self.inputs.items()}
        if self.mode == api.RecordMode.TRAIN or self.mode == api.RecordMode.VALIDATION:
            y = {k: _index_dict_to_arr(record, v) for k, v in self.outputs.items()}
            if self.sample_weights is not None:
                w = {k: np.array(record[v]) for k, v in self.sample_weights.items()}
                return (X, y, w)
            else:
                return (X, y)
        else:
            return (X,)



from typing import List

import tensorflow as tf

from barrage import config
from barrage.utils import import_utils


def build_network(cfg_model: dict, transform_params: dict) -> tf.keras.Model:
    """Build the network.

    Args:
        cfg_model: dict, model subsection of config.
        transform_params: dict, params from transformer.

    Returns:
        tf.keras.Model, network.

    Raises:
        TypeError, network not a tf.keras.Model.
    """
    path = cfg_model["network"]["import"]
    network_params = cfg_model["network"].get("params", {})

    net_func = import_utils.import_obj_with_search_modules(path)
    net = net_func(**network_params, **transform_params)

    if not isinstance(net, tf.keras.Model):
        raise TypeError(f"import network: {net} is not a tf.keras.Model")

    return net


def build_objective(cfg_model: dict) -> dict:
    """Build objective (loss, loss_weights, metrics, and sample_weight_mode)
    for each model output.

    Args:
        cfg_model: dict, model subsection of config.

    Returns:
        dict, objective
    """
    loss = {}
    loss_weights = {}
    sample_weight_mode = {}
    metrics = {}

    for output in cfg_model["outputs"]:
        name = output["name"]
        loss[name] = import_utils.import_loss(output["loss"])
        loss_weights[name] = output.get("loss_weight", 1.0)
        sample_weight_mode[name] = output.get("sample_weight_mode")
        metrics[name] = [
            import_utils.import_metric(m) for m in output.get("metrics", [])
        ]

    return {
        "loss": loss,
        "loss_weights": loss_weights,
        "sample_weight_mode": sample_weight_mode,
        "metrics": metrics,
    }


def check_output_names(cfg_model: dict, net: tf.keras.Model):
    """Check the net outputs in the config match the actual net.

    Args:
        cfg_model: dict, model subsection of config.
        net: tf.keras.Model, net.

    Raises:
        ValueError, mismatch between config and net.
    """
    config_net_outputs = {o["name"] for o in cfg_model["outputs"]}
    actual_net_outputs = set(net.output_names)
    if config_net_outputs != actual_net_outputs:
        raise ValueError(
            f"'config.model.outputs.names': {config_net_outputs} "
            f"mismatch actual model outputs: {actual_net_outputs} - "
            "order and names must exactly match"
        )


def sequential_from_config(layers: List[dict], **kwargs) -> tf.keras.Model:
    """Build a sequential model from a list of layer specifications.
    Supports references to network_params computed inside Transformers by specifying
    {{variable name}}.

    Args:
        layers: list[dict], layer imports.

    Returns:
        tf.keras.Model, network.
    """
    layers = config._render_params(layers, kwargs)
    network = tf.keras.models.Sequential()
    for layer in layers:

        if "import" not in layer:
            raise KeyError(f"layer {layer} missing 'import' key")
        if not layer.keys() <= {"import", "params"}:
            unexpected_keys = set(layer.keys()).difference({"import", "params"})
            raise KeyError(f"layer {layer} unexpected key(s): {unexpected_keys}")

        layer_cls = import_utils.import_obj_with_search_modules(
            layer["import"], search_modules=["tensorflow.keras.layers"]
        )
        layer_params = layer.get("params", {})
        network.add(layer_cls(**layer_params))

    return network



from tensorflow.python.keras import callbacks
from tensorflow.python.keras.optimizer_v2 import learning_rate_schedule, optimizer_v2

from barrage.utils import import_utils


def build_optimizer(cfg_solver: dict) -> optimizer_v2.OptimizerV2:
    """Build the optimizer.

    Args:
        cfg_solver: dict, solver subsection of config.

    Returns:
        optimizer_v2.OptimizerV2, tf.keras v2 optimizer.

    Raises:
        TypeError, optimizer not an OptimizerV2.
        TypeError, learning rate is not a float or LearningRateSchedule.
    """
    path = cfg_solver["optimizer"]["import"]
    params = cfg_solver["optimizer"].get("params", {})
    learning_rate = cfg_solver["optimizer"]["learning_rate"]

    if isinstance(learning_rate, dict):
        lr_cls = import_utils.import_obj_with_search_modules(
            learning_rate["import"],
            search_modules=["tensorflow.keras.optimizers.schedules"],
        )
        lr = lr_cls(**learning_rate["params"])
        if not isinstance(lr, learning_rate_schedule.LearningRateSchedule):
            raise TypeError(f"import learning rate: {lr} is not a LearningRateSchedule")
    else:
        lr = learning_rate

    opt_cls = import_utils.import_obj_with_search_modules(
        path, search_modules=["tensorflow.keras.optimizers"], search_both_cases=True
    )
    opt = opt_cls(learning_rate=lr, **params)

    if not isinstance(opt, optimizer_v2.OptimizerV2):
        raise TypeError(f"import optimizer: {opt} is not an OptimizerV2")

    return opt


def create_learning_rate_reducer(cfg_solver: dict) -> callbacks.ReduceLROnPlateau:
    """Create a ReduceLROnPlateau callback.

    Args:
        cfg_solver: dict, solver subsection of config.

    Returns:
        ReduceLROnPlateau, ReduceLROnPlateau callback.
    """
    params = cfg_solver["learning_rate_reducer"]
    params["verbose"] = 1

    return callbacks.ReduceLROnPlateau(**params)



import click

from barrage import BarrageModel
from barrage.utils import io_utils


@click.group(name="barrage")
def cli():
    pass


@cli.command()
@click.pass_context
def help(context):
    click.echo(context.parent.get_help())


@cli.command("train")
@click.argument("config", type=click.Path(exists=True))
@click.argument("train-data", type=click.Path(exists=True))
@click.argument("validation-data", type=click.Path(exists=True))
@click.option(
    "-a",
    "--artifact-dir",
    default="artifacts",
    type=click.Path(),
    help="location to save artifacts",
)
def train(config, train_data, validation_data, artifact_dir):
    """Barrage deep learning train.

    Supported filetypes:

        1. .csv

        2. .json

    Args:

        config: filepath to barrage config [REQUIRED].

        train-data: filepath to train data [REQUIRED].

        validation-data: filepath to validation data [REQUIRED].

    Note: artifact-dir cannot already exist.
    """
    cfg = io_utils.load_json(config)
    records_train = io_utils.load_data(train_data)
    records_validation = io_utils.load_data(train_data)
    BarrageModel(artifact_dir).train(cfg, records_train, records_validation)


@cli.command("predict")
@click.argument("score-data", type=click.Path(exists=True))
@click.argument("artifact-dir", type=click.Path(exists=True))
@click.option(
    "-o",
    "--output",
    default="scores.json",
    type=click.Path(),
    help="output filepath for scores",
)
def predict(score_data, artifact_dir, output):
    """Barrage deep learning predict.

    Supported filetypes:

        1. .csv

        2. .json

    Args:

        score-data: filepath to score data [REQUIRED].

        artifact-dir: location to load artifacts [REQUIRED].
    """
    records_score = io_utils.load_data(score_data)
    scores = BarrageModel(artifact_dir).predict(records_score)
    io_utils.save_json(scores, output)



from functools import partial
import importlib
import inspect
from typing import Callable, List, Union

from tensorflow.python.keras import losses, metrics

from barrage import logger


def import_obj_with_search_modules(
    python_path: str, search_modules: List[str] = None, search_both_cases=False
) -> Callable:
    """Import a object (variable, class, function, etc...) from a python path.
    search_modules are used to potentially shorten the python path by eliminating the
    module part of the string. Optionally check both uncapitalized and capitalized
    version of the object name.

    Args:
        python_path: str, full python path or object name (in conjunction with
             search_modules).
        search_modules: list[str], python modules against short path.
        search_both_cases: bool, try capitalized and uncapitalized object name.

    Returns:
        callable, imported object.

    Raises:
        ImportError, object not found.
    """
    python_path = _maybe_fix_tensorflow_python_path(python_path)
    try:
        module_path, obj_name = python_path.rsplit(".", 1)
    except ValueError:
        module_path = ""
        obj_name = python_path

    if module_path:
        module = importlib.import_module(module_path)
        if not hasattr(module, obj_name):
            raise ImportError(
                f"object: {obj_name} was not found in module {module_path}"
            )
        return getattr(module, obj_name)
    else:
        if search_modules is None:
            raise ImportError(f"object: {obj_name} not found, no module provided.")

        obj_names = _make_both_cases(obj_name) if search_both_cases else [obj_name]
        for search_module_path in search_modules:
            module = importlib.import_module(search_module_path)
            for name in obj_names:
                if hasattr(module, name):
                    return getattr(module, name)

        raise ImportError(
            f"object: {obj_name} was not found in the searched "
            f"modules {search_modules}"
        )


def import_or_alias(python_path: str) -> Union[Callable, str]:
    """Import an object from a full python path or assume it is an alias for some object
    in TensorFlow (e.g. "categorical_crossentropy").

    Args:
        python_path: str, full python path or alias.

    Returns:
        callable/str, import object/alias.

    Raises:
        ImportError, object not found.
    """
    python_path = _maybe_fix_tensorflow_python_path(python_path)
    try:
        module_path, obj_name = python_path.rsplit(".", 1)
    except ValueError:
        return python_path

    module = importlib.import_module(module_path)

    if not hasattr(module, obj_name):
        raise ImportError(f"object: {obj_name} was not found in module {module_path}")
    return getattr(module, obj_name)


def import_partial_wrap_func(import_block: dict) -> Callable:
    """Import a function from an import block and  wrap with partial.

    Args:
        import_block: dict, {"import": str, "params": dict}.

    Returns:
        function.
    """
    func = import_obj_with_search_modules(import_block["import"])
    params = import_block.get("params", {})
    return partial(func, **params)


def import_loss(import_block: dict) -> Union[losses.Loss, str]:
    """Import a loss from an import block.

    Args:
        import_block: dict, {"import": str, "params": dict}.

    Returns:
        Loss/str, non-alias loss or alias.

    Raises:
        TypeError, non-alias loss is not a Loss.
    """
    loss = import_or_alias(import_block["import"])
    if isinstance(loss, str):
        return loss
    else:
        params = import_block.get("params", {})
        if not inspect.isclass(loss):
            raise TypeError(
                f"import loss: {loss} must be a class for v2 tensorflow.keras API"
            )
        loss = loss(**params)
        if not isinstance(loss, losses.Loss):
            raise TypeError(
                f"import loss: {loss} is not a tensorflow.python.keras.losses.Loss"
            )
        return loss


def import_metric(import_block: dict) -> Union[metrics.Metric, losses.Loss, str]:
    """Import a metric from an import block.

    Note: A loss can be a metric, but a metric cannot always be a loss.

    Args:
        import_block: dict, {"import": str, "params": dict}.

    Returns:
        Metric/Loss/str, non-alias metric or alias.

    Raises:
        TypeError, non-alias metric is not a Metric/loss.
    """
    metric = import_or_alias(import_block["import"])
    if isinstance(metric, str):
        return metric
    else:
        params = import_block.get("params", {})
        if not inspect.isclass(metric):
            raise TypeError(
                f"import metric: {metric} must be a class "
                "for v2 tensorflow.keras API"
            )
        metric = metric(**params)
        if not isinstance(metric, (metrics.Metric, losses.Loss)):
            raise TypeError(
                f"import metric: {metric} is not a "
                "tensorflow.python.keras.metrics.Metric or "
                "tensorflow.python.keras.losses.Loss"
            )
        return metric


def _maybe_fix_tensorflow_python_path(python_path: str) -> str:
    """Fix a common mistake python path mistake tf.keras vs tensorflow.keras.

    Args:
        python_path: str, python path.

    Returns:
        str, python path.
    """
    if python_path.startswith("tf.keras"):
        python_path = python_path.replace("tf.keras", "tensorflow.keras")
        logger.warning(
            f"fixing import {python_path} - replacing tf.keras "
            "with tensorflow.python.keras"
        )
    return python_path


def _make_both_cases(s: str) -> List[str]:
    """Capitalize and uncapitalize first letter of string. Original case first.

    Args:
        s, str.

    Return:
        list[str], capitalized and uncapitalized version of string.
    """
    if s:
        upper = s[:1].upper() + s[1:]
        lower = s[:1].lower() + s[1:]
        if upper == s:
            return [upper, lower]
        else:
            return [lower, upper]
    else:
        return [""]



# Dataset
TRANSFORMER = {"import": "IdentityTransformer"}
AUGMENTOR = []  # type: list

# Solver
BATCH_SIZE = 32
EPOCHS = 50
OPTIMIZER = {"import": "Adam", "learning_rate": 1e-3}

# Services
BEST_CHECKPOINT = {"monitor": "val_loss", "mode": "min"}
TENSORBOARD = {}  # type: dict
TRAIN_EARLY_STOPPING = {
    "monitor": "loss",
    "mode": "min",
    "min_delta": 1e-5,
    "patience": 10,
    "verbose": 1,
}
VALIDATION_EARLY_STOPPING = {
    "monitor": "val_loss",
    "mode": "min",
    "min_delta": 1e-5,
    "patience": 10,
    "verbose": 1,
}



import json
import os
import pickle

import numpy as np
import pandas as pd


def load_json(filename: str, path: str = ""):
    """Load a json object.

    Args:
        filename: str.
        path: str.

    Returns:
        object.
    """
    with open(os.path.join(path, filename), "r") as fn:
        return json.load(fn)


def save_json(obj, filename: str, path: str = ""):
    """Save a json object.

    Args:
        obj: object.
        filename: str.
        path: str.
    """

    def _default(obj):
        if isinstance(obj, np.generic):
            return obj.item()
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            raise TypeError(f"invalid json serialization type: {type(obj)}")

    with open(os.path.join(path, filename), "w") as fn:
        json.dump(obj, fn, indent=2, sort_keys=True, default=_default)


def load_pickle(filename: str, path: str = ""):
    """Load a pickled object.

    Args:
        filename: str.
        path: str.

    Returns:
        object.
    """
    with open(os.path.join(path, filename), "rb") as fn:
        return pickle.load(fn)


def save_pickle(obj, filename: str, path: str = ""):
    """Save a pickled object.

    Args:
        obj: object.
        filename: str.
        path: str.
    """
    with open(os.path.join(path, filename), "wb") as fn:
        pickle.dump(obj, fn)


def load_data(filename: str, path: str = "") -> pd.DataFrame:
    """Load a file into a pandas DataFrame.

    Args:
        filename: str.
        path: str.

    Returns:
        pd.DataFrame.
    """
    filepath = os.path.join(path, filename)
    if not os.path.isfile(filepath):
        raise FileNotFoundError(f"filepath: {filepath} does not exist")

    ext = os.path.splitext(filepath)[1].lower()
    if ext == ".json":
        return pd.read_json(filepath)
    elif ext == ".csv":
        return pd.read_csv(filepath)
    else:
        raise ValueError("Unsupported extension: {ext}")



import copy
import os

import jsonschema

from barrage import defaults as d
from barrage.utils import io_utils


def prepare_config(cfg: dict) -> dict:
    """Prepare config for use - apply defaults, validate schema.

    Args:
        cfg: dict, config.

    Returns:
        dict, validated config with defaults.
    """
    cfg = _merge_defaults(cfg)
    _validate_schema(cfg)
    return cfg


def _merge_defaults(cfg: dict) -> dict:
    """Merge default params to config.

    Args:
        cfg: dict, config.

    Returns:
        dict, config with defaults.

    Raises:
        jsonschema.ValidationError: invalid config params.
    """
    cfg["dataset"] = cfg.get("dataset", {})
    if not isinstance(cfg["dataset"], dict):
        raise jsonschema.ValidationError("config param 'dataset' must be a dict")
    cfg["dataset"]["transformer"] = cfg["dataset"].get("transformer", d.TRANSFORMER)
    cfg["dataset"]["augmentor"] = cfg["dataset"].get("augmentor", d.AUGMENTOR)

    cfg["solver"] = cfg.get("solver", {})
    if not isinstance(cfg["solver"], dict):
        raise jsonschema.ValidationError("config param 'solver' must be a dict")

    cfg["solver"]["batch_size"] = cfg["solver"].get("batch_size", d.BATCH_SIZE)
    cfg["solver"]["epochs"] = cfg["solver"].get("epochs", d.EPOCHS)
    cfg["solver"]["optimizer"] = cfg["solver"].get("optimizer", d.OPTIMIZER)

    cfg["services"] = cfg.get("services", {})
    if not isinstance(cfg["services"], dict):
        raise jsonschema.ValidationError("config param 'services' must be a dict")

    cfg["services"]["best_checkpoint"] = cfg["services"].get(
        "best_checkpoint", d.BEST_CHECKPOINT
    )
    cfg["services"]["tensorboard"] = cfg["services"].get("tensorboard", d.TENSORBOARD)
    cfg["services"]["train_early_stopping"] = cfg["services"].get(
        "train_early_stopping", d.TRAIN_EARLY_STOPPING
    )
    cfg["services"]["validation_early_stopping"] = cfg["services"].get(
        "validation_early_stopping", d.VALIDATION_EARLY_STOPPING
    )

    return copy.deepcopy(cfg)


def _validate_schema(cfg: dict):
    """Validate config params against schema.

    Args:
        cfg: dict, config.

    Raises:
        jsonschema.ValidationError: invalid config params.
    """
    schema = io_utils.load_json(
        "schema.json", os.path.abspath(os.path.dirname(__file__))
    )
    try:
        jsonschema.validate(cfg, schema)
    except jsonschema.ValidationError as err:
        raise jsonschema.ValidationError(f"invalid barrage config: {err}")

    # Check model outputs have unique names
    num_outputs = len(cfg["model"]["outputs"])
    num_unique_names = len({o["name"] for o in cfg["model"]["outputs"]})
    if num_outputs != num_unique_names:
        raise jsonschema.ValidationError(
            f"invalid barrage config: 'outputs' names are not unique"
        )

    # Check that multi-output networks have loss weights for each output
    if num_outputs > 1:
        if not all("loss_weight" in output for output in cfg["model"]["outputs"]):
            raise jsonschema.ValidationError(
                "invalid barrage config: each output in 'outputs' requires a "
                "'loss_weight' for multi output networks"
            )


def _render_params(cfg, params: dict):  # noqa C:901
    """Render a config or config section with params jinja style.

    Args:
        cfg: config.
        params: dict, render params.

    Returns:
        config.
    """

    def _replace_item(obj, mapping):

        if isinstance(obj, dict):
            for k, v in obj.items():
                if isinstance(v, (list, dict)):
                    obj[k] = _replace_item(v, mapping)
                elif isinstance(v, str) and v in mapping:
                    obj[k] = mapping[v]
        elif isinstance(obj, list):
            for k, v in enumerate(obj):
                if isinstance(v, (list, dict)):
                    obj[k] = _replace_item(v, mapping)
                elif isinstance(v, str) and v in mapping:
                    obj[k] = mapping[v]
        return obj

    # jinja style
    mapping = {"{{" + k + "}}": v for k, v in params.items()}
    cfg = _replace_item(cfg, mapping)

    return cfg



import os
from typing import List

from tensorflow.python.keras import callbacks

from barrage import logger

DATASET = "dataset"
BEST_CHECKPOINT = "best_checkpoint"
BEST_MODEL = "model_best.ckpt"
RESUME_CHECKPOINTS = "resume_checkpoints"
RESUME_MODEL = "model_epoch_{epoch:04d}.ckpt"
TENSORBOARD = "TensorBoard"
CSV_LOGGER_FILENAME = "training_report.csv"
REQUIRED_SUBDIRS = [DATASET, BEST_CHECKPOINT, RESUME_CHECKPOINTS, TENSORBOARD]


def make_artifact_dir(artifact_dir: str):
    """Make the artifact directory and all required subdirectories.

    Args:
        artifact_dir: str, path to artifact directory.

    Raises:
        OSError, artifact_dir already exists.
    """
    if os.path.isdir(artifact_dir):
        raise OSError(f"artifact_dir: {artifact_dir} already exists")

    os.mkdir(artifact_dir)
    for subdir in REQUIRED_SUBDIRS:
        os.mkdir(os.path.join(artifact_dir, subdir))


def create_all_services(
    artifact_dir: str, cfg_services: dict
) -> List[callbacks.Callback]:
    """Create all services (callbacks).

    Args:
        artifact_dir: str, path to artifact directory.
        cfg_services: dict, services subsection of config.

    Returns:
        list[Callback], all services.
    """
    return [
        _create_best_checkpoint(artifact_dir, cfg_services),
        _create_resume_checkpoint(artifact_dir),
        _create_tensorboard(artifact_dir, cfg_services),
        _create_csv_logger(artifact_dir),
        _create_train_early_stopping(cfg_services),
        _create_validation_early_stopping(cfg_services),
        callbacks.TerminateOnNaN(),
    ]


def _create_best_checkpoint(
    artifact_dir: str, cfg_services: dict
) -> callbacks.ModelCheckpoint:
    """Create a callback that saves the best model.

    Args:
        artifact_dir: str, path to artifact directory.
        cfg_services: dict, services subsection of config.

    Returns:
        ModelCheckpoint, callback that saves the best model.
    """
    checkpoint_params = cfg_services["best_checkpoint"]
    checkpoint_params["monitor"] = _force_monitor_to_mode(
        checkpoint_params["monitor"], True, "best_checkpoint"
    )
    filepath = get_best_checkpoint_filepath(artifact_dir)
    return callbacks.ModelCheckpoint(
        filepath=filepath,
        monitor=checkpoint_params["monitor"],
        mode=checkpoint_params["mode"],
        verbose=1,
        save_best_only=True,
        save_weights_only=True,
    )


def _create_resume_checkpoint(artifact_dir: str) -> callbacks.ModelCheckpoint:
    """Create a callback that saves the model every epoch.

    Args:
        artifact_dir: str, path to artifact directory.

    Returns:
        ModelCheckpoint, callback that saves the model every epoch.
    """
    filepath = get_resume_checkpoints_filepath(artifact_dir)
    return callbacks.ModelCheckpoint(
        filepath=filepath,
        monitor="val_loss",
        mode="min",
        verbose=0,
        save_best_only=False,
        save_weights_only=True,
    )


def _create_tensorboard(artifact_dir: str, cfg_services: dict) -> callbacks.TensorBoard:
    """Create a TensorBoard callback.

    Args:
        artifact_dir: str, path to artifact directory.
        cfg_services: dict, services subsection of config.

    Returns:
        TensorBoard, Tensorboard callback.
    """
    tensorboard_params = cfg_services["tensorboard"]
    if "log_dir" in tensorboard_params:
        logger.warning("'log_dir' automatically handled for 'tensorboard' service")
    tensorboard_params["log_dir"] = os.path.join(artifact_dir, TENSORBOARD)
    return callbacks.TensorBoard(**tensorboard_params)


def _create_csv_logger(artifact_dir: str) -> callbacks.CSVLogger:
    """Create a CSVLogger callback.

    Args:
        artifact_dir: str, path to artifact directory.

    Returns:
        CSVLogger, CSVLogger callbackk.
    """
    filename = os.path.join(artifact_dir, CSV_LOGGER_FILENAME)
    return callbacks.CSVLogger(filename=filename, separator=",", append=True)


def _create_train_early_stopping(cfg_services: dict,) -> callbacks.EarlyStopping:
    """Create an early stopping callback that monitors a training 'metric'.

    Args:
        cfg_services: dict, services subsection of config.

    Returns:
        EarlyStopping, EarlyStopping callback that monitors a training 'metric'.
    """
    early_stopping_params = cfg_services["train_early_stopping"]
    early_stopping_params["monitor"] = _force_monitor_to_mode(
        early_stopping_params["monitor"], False, "train_early_stopping"
    )
    return callbacks.EarlyStopping(**early_stopping_params)


def _create_validation_early_stopping(cfg_services: dict) -> callbacks.EarlyStopping:
    """Create an early stopping callback that monitors a validation 'metric'.

    Args:
        cfg_services: dict, services subsection of config.

    Returns:
        EarlyStopping, EarlyStopping callback that monitors a validation 'metric'.
    """
    early_stopping_params = cfg_services["validation_early_stopping"]
    early_stopping_params["monitor"] = _force_monitor_to_mode(
        early_stopping_params["monitor"], True, "validation_early_stopping"
    )
    return callbacks.EarlyStopping(**early_stopping_params)


def get_best_checkpoint_filepath(artifact_dir: str) -> str:
    """Get the filepath for the best checkpoint.

    Args:
        artifact_dir: str, path to artifact directory.

    Returns:
        str, filepath for best checkpoint directory.
    """
    return os.path.join(artifact_dir, BEST_CHECKPOINT, BEST_MODEL)


def get_resume_checkpoints_filepath(artifact_dir: str) -> str:
    """Get the filepath for the resume checkpoints.

    Args:
        artifact_dir: str, path to artifact directory.

    Returns:
        str, filepath for resume checkpoints.
    """
    return os.path.join(artifact_dir, RESUME_CHECKPOINTS, RESUME_MODEL)


def _force_monitor_to_mode(monitor: str, to_val: bool, service_name: str) -> str:
    """Force a monitor quantity to either train or validation mode. For
    example 'loss' - train, 'val_loss' - validation.

    Args:
        monitor: str, metric to monitor.
        to_val: bool, validation if true, else false.
        service_name: str, corresponding service (for warning purposes).

    Returns:
        str, monitor maybe forced.
    """
    if to_val and not monitor.startswith("val_"):
        monitor = f"val_{monitor}"
        logger.warning(
            f"corrected 'monitor' to validation verison: {monitor} "
            f"for service: {service_name}"
        )
    elif not to_val and monitor.startswith("val_"):
        monitor = monitor[4:]
        logger.warning(
            f"corrected 'monitor' to train verison: {monitor} "
            f"for service: {service_name}"
        )
    return monitor



import abc
import enum
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import pandas as pd

Record = Dict[str, Any]
Records = List[Record]
InputRecords = Union[Records, pd.DataFrame]

DataRecord = Tuple[Dict[str, Union[np.ndarray, float]], ...]
BatchDataRecords = Tuple[Dict[str, np.ndarray], ...]

RecordScore = Dict[str, np.ndarray]
BatchRecordScores = List[RecordScore]


class RecordMode(enum.Enum):
    TRAIN = 0
    VALIDATION = 1
    SCORE = 2


class RecordLoader(abc.ABC):
    """Class for loading records into DataRecord.

    Args:
        mode: RecordMode, load mode.
    """

    def __init__(self, mode: RecordMode, **params):
        self.mode = mode

    def __call__(self, record: Record) -> DataRecord:
        return self.load(record)

    @abc.abstractmethod
    def load(self, record: Record) -> DataRecord:  # pragma: no cover
        """Method for loading a record into DataRecord.

        Args:
            record: Record, record.

        Returns:
            DataRecord, data record.
        """
        raise NotImplementedError()


class RecordTransformer(abc.ABC):
    """Class that computes a transform on training data records & applys
    transform to validation and scoring data records (network input), ability
    to pass computed network params to the network builder, and
    ability to apply inverse transforms on record scores (network output).

    Args:
        mode: RecordMode, transform mode.
        loader: RecordLoader, record loader.
    """

    def __init__(self, mode: RecordMode, loader: RecordLoader, **params):
        self.mode = mode
        self.loader = loader
        self._network_params = {}  # type: dict

    @abc.abstractmethod
    def fit(self, records: Records):  # pragma: no cover
        """Fit transform to records.

        Args:
            records: Records, records.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def transform(self, data_record: DataRecord) -> DataRecord:  # pragma: no cover
        """Apply transform to a data record.

        Args:
            data_record: DataRecord, data record.

        Returns:
            DataRecord, data record.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def postprocess(self, score: RecordScore) -> RecordScore:  # pragma: no cover
        """Postprocess score to undo transform.

        Args:
            score: RecordScore, record output from net.

        Returns:
            RecordScore, postprocessed record output from net.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def load(self, path: str):  # pragma: no cover
        """Load transformer.

        Args:
            path: str.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def save(self, path: str):  # pragma: no cover
        """Save transformer.

        Args:
            path: str.
        """
        raise NotImplementedError()

    @property
    def network_params(self) -> dict:
        """Special params passed to the network builder."""
        return self._network_params

    @network_params.setter
    def network_params(self, x):
        if not isinstance(x, dict):
            raise TypeError("network_params must be a dict")
        self._network_params = x



from barrage import api


class IdentityTransformer(api.RecordTransformer):
    """Default transformer that does nothing (identity transform) that ensures
    every dataset has a transformer.
    """

    def fit(self, records: api.Records):
        """Pass - no transform to fit.

        Args:
            records: Records, records.
        """
        pass

    def transform(self, data_record: api.DataRecord) -> api.DataRecord:
        """Identity - return the original data record unchanged.

        Args:
            data_record: DataRecord, data record.

        Returns:
            DataRecord, data record.
        """
        return data_record

    def postprocess(self, score: api.RecordScore) -> api.RecordScore:
        """Identity - return the record score unchanged.

        Args:
            score: RecordScore, record output from net.

        Returns:
            RecordScore, record output from net.
        """
        return score

    def save(self, path: str):
        """Pass - no objects to save.

        Args:
            path: str.
        """
        pass

    def load(self, path: str):
        """Pass - no objects to load.

        Args:
            path: str.
        """
        pass



import copy
import os
from typing import List, Union

import cytoolz
import numpy as np
import pandas as pd
import tensorflow as tf

from barrage import api, logger
from barrage.utils import import_utils

SEARCH_MODULES = ["barrage.dataset"]


class RecordDataset(tf.keras.utils.Sequence):
    """A sequence designed to wrap a DataFrame-like object: apply load
    operations, fit & apply transforms, apply data augmentation, and support sampling
    of records.

    At train time: fit transform -> batch -> load -> transform -> augment.

    At validation time: batch -> load -> transform.

    At score time: batch -> load -> transform.

    Args:
        artifact_dir: str, path to artifact directory.
        cfg_dataset: dict, dataset subsection of config.
        records: api.InputRecords, records.
        mode: RecordMode, transform mode.
        batch_size: int, batch size.
    """

    def __init__(
        self,
        artifact_dir: str,
        cfg_dataset: dict,
        records: api.InputRecords,
        mode: api.RecordMode,
        batch_size: int,
    ):

        if not isinstance(mode, api.RecordMode):
            raise TypeError("mode must be type RecordMode")

        # Standardize InputRecords to Records
        if isinstance(records, pd.DataFrame):
            records.reset_index(drop=True, inplace=True)
            self.records = records.to_dict(orient="records")
        elif all(isinstance(record, dict) for record in records):
            self.records = records
        else:
            raise TypeError("record must be a list of dicts or pandas DataFrame")

        self.num_records = len(records)
        logger.info(f"Building {mode} dataset with {self.num_records} records")
        self.mode = mode
        self.batch_size = batch_size

        self.seed = cfg_dataset.get("seed")
        np.random.seed(self.seed)

        sample_count = cfg_dataset.get("sample_count")
        if self.mode == api.RecordMode.TRAIN and sample_count is not None:
            self._sample_inds = self.convert_sample_count_to_inds(
                [record[sample_count] for record in self.records]
            )
        else:
            self._sample_inds = list(range(self.num_records))
        self.shuffle()

        loader_cls = import_utils.import_obj_with_search_modules(
            cfg_dataset["loader"]["import"], search_modules=SEARCH_MODULES
        )
        self.loader = loader_cls(mode=mode, **cfg_dataset["loader"].get("params", {}))
        if not isinstance(self.loader, api.RecordLoader):
            raise TypeError(f"loader {self.loader} is not of type RecordLoader")

        transformer_cls = import_utils.import_obj_with_search_modules(
            cfg_dataset["transformer"]["import"], search_modules=SEARCH_MODULES
        )
        self.transformer = transformer_cls(
            mode=self.mode,
            loader=self.loader,
            **cfg_dataset["transformer"].get("params", {}),
        )
        if not isinstance(self.transformer, api.RecordTransformer):
            raise TypeError(
                f"transformer {self.transformer} is not of type RecordTransformer"
            )

        dataset_dir = os.path.join(artifact_dir, "dataset")
        if self.mode == api.RecordMode.TRAIN:
            self.augmentor = RecordAugmentor(cfg_dataset["augmentor"])
            logger.info(f"Fitting transform: {self.transformer.__class__.__name__}")
            self.transformer.fit(copy.deepcopy(self.records))
            logger.info(
                f"Transformer network params: {self.transformer.network_params}"
            )
            self.transformer.save(dataset_dir)
        else:
            self.transformer.load(dataset_dir)

    def __len__(self):
        """Number of batches in a sequence."""
        return int(np.ceil(len(self.sample_inds) / float(self.batch_size)))

    def __getitem__(self, ind) -> api.BatchDataRecords:
        """Get a batch by index.

        Args:
            ind: int, batch index.

        Returns:
            BatchDataRecords, batch data records.
        """
        batch_inds = self.sample_order[
            ind * self.batch_size : (ind + 1) * self.batch_size
        ]
        batch_records = copy.deepcopy([self.records[bi] for bi in batch_inds])

        if self.mode == api.RecordMode.TRAIN:
            lst_data_records = [
                self.augmentor(self.transformer.transform(self.loader(record)))
                for record in batch_records
            ]
        else:
            lst_data_records = [
                self.transformer.transform(self.loader(record))
                for record in batch_records
            ]

        return batchify_data_records(lst_data_records)

    def on_epoch_end(self):
        """Shuffle sample_order on epoch end."""
        self.shuffle()

    @property
    def sample_inds(self):
        """Index of records used for batches. Contains repeats if
        self.mode = RecordMode.TRAIN and self.sample_count.
        """
        return self._sample_inds.copy()

    @sample_inds.setter
    def sample_inds(self, x):
        raise ValueError("illegal set on 'sample_inds'")

    def shuffle(self):
        """Shuffle sample_inds to compute sample_order."""
        if self.mode == api.RecordMode.TRAIN or self.mode == api.RecordMode.VALIDATION:
            self.sample_order = self.sample_inds
            np.random.shuffle(self.sample_order)
        else:
            self.sample_order = self.sample_inds

    @staticmethod
    def convert_sample_count_to_inds(sample_count: List[int]) -> List[int]:
        """Convert a list of sample counts to a list of sample inds.

        Args:
            sample_count: list[int], integer list of sample counts.

        Returns:
            list[int], sample inds.
        """
        sample_count = [max(1, int(sc)) for sc in sample_count]
        sample_lsts = [[ind] * val for ind, val in enumerate(sample_count)]
        sample_inds = [item for li in sample_lsts for item in li]
        return sample_inds


class RecordAugmentor(object):
    """Class for applying a list of data augmentation functions to a data record.

    Args:
        funcs: list[dict], list of augmentation functions
            {"import": "python_path", "params": {...}}.
    """

    def __init__(self, funcs: List[dict]):
        self.augment_func = self.reduce_compose(
            *[import_utils.import_partial_wrap_func(f) for f in funcs]
        )

    def __call__(self, data_record: api.DataRecord) -> api.DataRecord:
        return self.augment(data_record)

    def augment(self, data_record: api.DataRecord) -> api.DataRecord:
        """Apply augmentation to a train data record.

        Args:
            data_record: DataRecord, data record.

        Returns:
            DataRecord, augmented data record.
        """
        return self.augment_func(data_record)

    @staticmethod
    def reduce_compose(*funcs):
        """Compose a list of functions into a single function."""
        if len(funcs) == 0:
            return lambda x: x

        from functools import reduce

        def _compose2(func1, func2):
            return lambda *args, **kwargs: func2(func1(*args, **kwargs))

        return reduce(_compose2, funcs)


def batchify_data_records(data_records: List[api.DataRecord]) -> api.BatchDataRecords:
    """Stack a list of DataRecord into BatchRecord. This process converts a list
    of tuples comprising of dicts {str: float/array} into tuples of dict {str: array}.
    Float/array is concatenated along the first dimension. See Example.

    Args:
        data_records: list[DataRecord], list of individual data records.

    Returns:
        BatchDataRecords, batch data records.

    Example:
    ::

        data_record_1 = ({"input_1": 1, "input_2": 2}, {"output_1": 3})
        data_record_2 = ({"input_1": 2, "input_2": 4}, {"output_1": 6})
        batch_data_records = (
            {"input_1": arr([1, 2], "input_2": arr([2, 4])},
            {"output_1": arr([3, 6])}
        )
    """
    batch_data_records = tuple(
        cytoolz.merge_with(np.array, ii) for ii in zip(*data_records)
    )
    return batch_data_records  # type: ignore


def batchify_network_output(
    network_output: Union[np.ndarray, List[np.ndarray]], output_names: List[str]
) -> api.BatchRecordScores:
    """Convert network output scores to BatchRecordScores. This process converts a
    single numpy array or list of numpy arrays into a list of dictionaries. See example.

    Args:
        network_output: union[np.ndarray, list[np.ndarray], network output.

    Returns:
        BatchRecordScores, batch scores.

    Example:
    ::

        network_output == np.array([[1], [2]])
        output_names = ["y"]
        batch_scores = [{"y": np.array([1])}, {"y": np.array([2])}]
    """
    # Handle type inconsistency between outputs of single output/multi networks
    if isinstance(network_output, np.ndarray):
        dict_output = {output_names[0]: network_output}
        num_scores = len(network_output)
    else:
        dict_output = {
            output_names[ii]: network_output[ii] for ii in range(len(output_names))
        }
        num_scores = len(network_output[0])

    scores = [
        {k: v[ii, ...] for k, v in dict_output.items()}  # type: ignore
        for ii in range(num_scores)
    ]

    return scores



from setuptools import find_packages, setup

long_description = """
Barrage is an opinionated supervised deep learning tool built on top of
TensorFlow 2.x designed to standardize and orchestrate the training and scoring of
complicated models. Barrage is built around a JSON config and the
TensorFlow 2.x library using the Tensorflow.Keras API.
"""

setup(
    name="barrage",
    version="0.4.0",
    description="A supervised deep learning tool.",
    long_description=long_description,
    author="Brian Nemsick",
    author_email="brian.nemsick@gmail.com",
    url="https://github.com/briannemsick/barrage/",
    license="MIT",
    python_requires=">=3.6",
    install_requires=[
        "absl-py>=0.8.0",  # <0.8.0 conflicts with logging module
        "click",
        "cytoolz>=0.9.0",
        "jsonschema>=3.0",
        "numpy",
        "pandas",
        "tensorflow>=2.2.0, <2.3",
    ],
    extras_require={
        "tests": [
            "black",
            "codecov",
            "flake8",
            "flake8-bugbear",
            "flake8-comprehensions",
            "flake8-print",
            "isort",
            "mypy",
            "pre-commit",
            "pre-commit-hooks",
            "pytest",
            "pytest-cov",
        ]
    },
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    entry_points={"console_scripts": ["barrage = barrage.console:cli"]},
    include_package_data=True,
    packages=find_packages(exclude=["docs*", "examples*", "tests*"]),
)



from .core import (  # noqa
    RecordAugmentor,
    RecordDataset,
    batchify_data_records,
    batchify_network_output,
)
from .loader import KeySelector  # noqa
from .transformer import IdentityTransformer  # noqa



import tensorflow as tf

from barrage import api, config, dataset, logger, model, services, solver
from barrage.utils import io_utils, tf_utils


class BarrageModel(object):
    """Class for training the network and scoring records with best performing
    network.

    Args:
        artifact_dir: str, path to artifact directory.
    """

    def __init__(self, artifact_dir):
        self._artifact_dir = artifact_dir

    def train(
        self,
        cfg: dict,
        records_train: api.InputRecords,
        records_validation: api.InputRecords,
    ) -> tf.keras.Model:
        """Train the network.

        Args:
            cfg: dict, config.
            records_train: InputRecords, training records.
            records_validation: InputRecords, validation records.

        Returns:
            tf.keras.Model, trained network.
        """
        logger.info("Starting training")
        tf_utils.reset()
        cfg = config.prepare_config(cfg)

        logger.info(f"Creating artifact directory: {self.artifact_dir}")
        services.make_artifact_dir(self.artifact_dir)
        io_utils.save_json(cfg, "config.json", self.artifact_dir)
        io_utils.save_pickle(cfg, "config.pkl", self.artifact_dir)

        logger.info("Creating datasets")
        ds_train = dataset.RecordDataset(
            artifact_dir=self.artifact_dir,
            cfg_dataset=cfg["dataset"],
            records=records_train,
            mode=api.RecordMode.TRAIN,
            batch_size=cfg["solver"]["batch_size"],
        )
        ds_validation = dataset.RecordDataset(
            artifact_dir=self.artifact_dir,
            cfg_dataset=cfg["dataset"],
            records=records_validation,
            mode=api.RecordMode.VALIDATION,
            batch_size=cfg["solver"]["batch_size"],
        )
        network_params = ds_train.transformer.network_params
        io_utils.save_json(network_params, "network_params.json", self.artifact_dir)
        io_utils.save_pickle(network_params, "network_params.pkl", self.artifact_dir)

        logger.info("Building network")
        net = model.build_network(cfg["model"], network_params)
        model.check_output_names(cfg["model"], net)

        logger.info("Compiling network")
        opt = solver.build_optimizer(cfg["solver"])
        objective = model.build_objective(cfg["model"])
        net.compile(optimizer=opt, **objective)

        logger.info("Creating services")
        callbacks = services.create_all_services(self.artifact_dir, cfg["services"])

        if "learning_rate_reducer" in cfg["solver"]:
            logger.info("Creating learning rate reducer")
            callbacks.append(solver.create_learning_rate_reducer(cfg["solver"]))

        logger.info("Training network")
        net.summary()
        net.fit(
            ds_train,
            validation_data=ds_validation,
            epochs=cfg["solver"]["epochs"],
            steps_per_epoch=cfg["solver"].get("steps"),
            callbacks=callbacks,
            verbose=1,
        )

        return net

    def predict(self, records_score: api.InputRecords) -> api.BatchRecordScores:
        """Score records.

        Args:
            records_score: InputRecords, scoring records.

        Returns:
            BatchRecordScores, scored data records.
        """
        if not hasattr(self, "net"):
            self.load()

        ds_score = dataset.RecordDataset(
            artifact_dir=self.artifact_dir,
            cfg_dataset=self.cfg["dataset"],
            records=records_score,
            mode=api.RecordMode.SCORE,
            batch_size=self.cfg["solver"]["batch_size"],
        )

        network_output = self.net.predict(ds_score, verbose=1)
        scores = [
            ds_score.transformer.postprocess(score)
            for score in dataset.batchify_network_output(
                network_output, self.net.output_names
            )
        ]

        return scores

    def load(self):
        """Load the best performing checkpoint."""

        # Load artifacts needed to recreate the network
        self.cfg = io_utils.load_pickle("config.pkl", self.artifact_dir)
        network_params = io_utils.load_pickle("network_params.pkl", self.artifact_dir)

        # Build network
        self.net = model.build_network(self.cfg["model"], network_params)

        # Load best checkpoint
        path = services.get_best_checkpoint_filepath(self.artifact_dir)
        self.net.load_weights(path).expect_partial()  # not loading optimizer

        return self

    @property
    def artifact_dir(self):
        return self._artifact_dir



import logging

logger = logging.getLogger(__name__)

logger.setLevel(logging.INFO)

console = logging.StreamHandler()
formatter = logging.Formatter(
    "Barrage %(asctime)s %(levelname)-5.4s: %(message)s", datefmt="%m/%d/%y %I:%M:%S"
)
console.setFormatter(formatter)
logger.addHandler(console)
